@@ -1,5 +1,4 @@
 /* ACME
- * init_tray (acme);
  * Copyright (C) 2001 Bastien Nocera <hadess@hadess.net>
  *
  * acme.c
@@ -30,10 +29,10 @@
 #include <X11/XF86keysym.h>
 /* Gnome headers */
 #include <gdk/gdkx.h>
-#include <gtk/gtkinvisible.h>
 #include <gnome.h>
 #include <glade/glade.h>
 #include <gconf/gconf-client.h>
+#include <libwnck/libwnck.h>
 #include "eggtrayicon.h"
 
 #include "acme.h"
@@ -42,9 +41,6 @@
 #include "fb-level.h"
 #endif
 
-static gboolean event_cb (GtkWidget * widget, GdkEvent * event,
-			  gpointer data);
-
 typedef struct {
 	GVolume *volobj;
 #ifdef USE_FBLEVEL
@@ -52,10 +48,10 @@ typedef struct {
 #endif
 	GladeXML *xml;
 	GtkWidget *dialog;
-	GtkWidget *invisible;
 	int NumLockMask, CapsLockMask, ScrollLockMask;
 	GConfClient *conf_client;
 	guint dialog_timeout;
+	WnckScreen *screen;
 
 	/* Tray icon */
 	EggTrayIcon *tray_icon;
@@ -70,6 +66,8 @@ enum {
 	ICON_EJECT,
 };
 
+static void init_tray (Acme *acme);
+
 static void
 selection_get_func (GtkClipboard *clipboard, GtkSelectionData *selection_data,
 		guint info, gpointer user_data_or_owner)
@@ -178,7 +176,7 @@ execute_this_or_that (char *cmd1, char *cmd2)
 				(_("Couldn't execute either command: %s\n"
 				   "or command: %s\n"
 				   "Verify that at least one of these commands"
-				   "exist."),
+				   " exist."),
 				 cmd1, cmd2);
 
 			acme_error (msg);
@@ -401,18 +399,9 @@ update_kbd_cb (GConfClient *client, guint id, GConfEntry *entry, gpointer data)
 static gboolean
 init_kbd (Acme *acme)
 {
-	GtkWidget *widget = acme->invisible;
 	int i;
 	gboolean retval = FALSE;
 
-	gtk_widget_show (widget);
-
-	/* Make the root window send events to the invisible proxy widget */
-	gdk_window_set_user_data (gdk_get_default_root_window (), widget);
-
-	gdk_window_set_events (gdk_get_default_root_window (),
-			GDK_KEY_PRESS_MASK);
-
 	for (i = 0; i < HANDLED_KEYS; i++)
 	{
 		int tmp;
@@ -512,6 +501,13 @@ tray_icon_press (GtkWidget *widget, GdkEventButton *event, Acme *acme)
 	return FALSE;
 }
 
+static gboolean
+tray_destroyed (GtkWidget *widget, GdkEvent *event, gpointer user_data)
+{
+	init_tray ((Acme *) user_data);
+	return TRUE;
+}
+
 static void
 init_tray (Acme *acme)
 {
@@ -552,6 +548,9 @@ init_tray (Acme *acme)
 	gtk_container_add (GTK_CONTAINER (evbox), image);
 	gtk_container_add (GTK_CONTAINER (acme->tray_icon), evbox);
 	gtk_widget_show_all (GTK_WIDGET (acme->tray_icon));
+
+	g_signal_connect (G_OBJECT (acme->tray_icon), "destroy-event",
+			G_CALLBACK (tray_destroyed), (gpointer) acme);
 }
 
 static void
@@ -613,6 +612,63 @@ dialog_show (Acme *acme)
 			(GtkFunction) dialog_hide, acme);
 }
 
+static void
+do_close_window_action (Acme *acme)
+{
+	GList *windows, *item;
+	WnckWindow *focused;
+
+	focused = NULL;
+	windows = wnck_screen_get_windows (acme->screen);
+
+	if (windows == NULL)
+		return;
+
+	for (item = windows; item != NULL; item = item->next)
+	{
+		WnckWindow *window = (WnckWindow *) (item->data);
+		if (wnck_window_is_active (window) == TRUE)
+		{
+			focused = window;
+			break;
+		}
+	}
+
+	if (focused != NULL)
+		wnck_window_close (focused);
+}
+
+static void
+do_shade_window_action (Acme *acme)
+{
+	GList *windows, *item;
+	WnckWindow *focused;
+
+	focused = NULL;
+	windows = wnck_screen_get_windows (acme->screen);
+
+	if (windows == NULL)
+		return;
+
+	for (item = windows; item != NULL; item = item->next)
+	{
+		WnckWindow *window = (WnckWindow *) (item->data);
+		if (wnck_window_is_active (window) == TRUE)
+		{
+			focused = window;
+			break;
+		}
+	}
+
+	if (focused != NULL)
+	{
+		if (wnck_window_is_shaded (focused) == TRUE)
+			wnck_window_unshade (focused);
+		else
+			wnck_window_shade (focused);
+	}
+}
+
 static void
 do_www_action (Acme *acme)
 {
@@ -814,6 +870,12 @@ do_action (int type, Acme *acme)
 	case RECORD_KEY:
 		execute ("gnome-sound-recorder", FALSE);
 		break;
+	case CLOSE_WINDOW_KEY:
+		do_close_window_action (acme);
+		break;
+	case SHADE_WINDOW_KEY:
+		do_shade_window_action (acme);
+		break;
 #ifdef USE_FBLEVEL
 	case BRIGHT_DOWN_KEY:
 	case BRIGHT_UP_KEY:
@@ -826,29 +888,33 @@ do_action (int type, Acme *acme)
 	}
 }
 
-static gboolean
-event_cb (GtkWidget * widget, GdkEvent * event, gpointer data)
+static GdkFilterReturn
+acme_filter_events (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 {
 	Acme *acme = (Acme *) data;
-	GdkEventKey *key;
+	XEvent *xev = (XEvent *) xevent;
+	XKeyEvent *key = (XKeyEvent *) xevent;
 	int i;
 
-	if (event->type != GDK_KEY_RELEASE)
-		return TRUE;
+	if (xev->type != KeyRelease)
+		return GDK_FILTER_CONTINUE;
 
-	key = (GdkEventKey *) event;
+	key = (XKeyEvent *) xevent;
 
-	for (i = 0; i < HANDLED_KEYS; i++) {
+	for (i = 0; i < HANDLED_KEYS; i++)
+	{
 #ifdef DEBUG
 		g_print ("comparing %d and %d\n", keys[i].key_code,
-				key->hardware_keycode);
+				key->keycode);
 #endif
-		if (keys[i].key_code == key->hardware_keycode) {
+		if (keys[i].key_code == key->keycode)
+		{
 			do_action (keys[i].key_type, acme);
-			break;
+			return GDK_FILTER_REMOVE;
 		}
 	}
-	return TRUE;
+
+	return GDK_FILTER_CONTINUE;
 }
 
 int
@@ -893,7 +959,6 @@ main (int argc, char *argv[])
 			GCONF_CLIENT_PRELOAD_ONELEVEL,
 			NULL);
 
-	acme->invisible = gtk_invisible_new ();
 	if (init_kbd (acme) == FALSE)
 	{
 		execute ("acme-properties", FALSE);
@@ -901,6 +966,7 @@ main (int argc, char *argv[])
 	}
 	init_sm (acme);
 	init_tray (acme);
+	acme->screen = wnck_screen_get_default ();
 	gtk_widget_realize (acme->dialog);
 	acme->dialog_timeout = 0;
 
@@ -928,9 +994,8 @@ main (int argc, char *argv[])
 		fb_problem_cb ();
 #endif
 
-	g_signal_connect (GTK_OBJECT (acme->invisible),
-			    "event",
-			    G_CALLBACK (event_cb), acme);
+	/* Start filtering the events */
+	gdk_window_add_filter (NULL, acme_filter_events, (gpointer) acme);
 
 	gtk_main ();
 
